/* ----------------------------------------------------
                          uniface3.c

   "universelles" Interface zum AVR-Controller zum
   "Spielen" ueber die serielle Schnittstelle.

   Integrierter ISP-Flasher fuer MCS-51
       -89S51
       -89S52
       -89S2051
       -89S4051
       -89S8253

   Nach Einschalten ist Benutzung als MCS-51 Flasher
   aktiv, um in den "SPI Experimentiermodus" zu gelangen
   ist ein Terminalprogramm notwendig:

        115200 Baud, 8N1

   Druecken der Taste 5 startet den SPI Kommandomodus,
   Exit beendet diesen wieder.

   Pinbelegungen, die Anschluesse vom AVR zum SPI-Zielgeraet
   sind ueber 100 Reihenwiderstaende (zur Bedaempfung) an-
   zuschliesen:

       SPI_MOSI_PIN    PB3           Arduino digital 11
       SPI_MISO_PIN    PB4                "          12
       SPI_CLK_PIN     PB5                "          13
       SPI_RST_PIN     PB2                "          10


   Zweck: Erleichterung von Experimenten mit ISP / SPI


   2014   R. Seelig

   ---------------------------------------------------- */


#include <avr/io.h>
#include <avr/pgmspace.h>
#include <util/delay.h>

#include "spi_uart.h"

#define page8253     64
#define page2051     32

char     rinv      = 0;
char     clkinv    = 0;
char     aktfreq   = 3;
char     mcutyp    = 0xff;
int      pagesize;

#define p_delay   4700          // Delay pagewrite in uS
#define p_delay2  4700          // Delay pagewrite in uS
#define b_delay     50          // Delay bytewrite in uS
#define b_delay2  1500          // Delay bytewrite in uS (80S2051 / S4051)
#define e_delay    420          // Delay erase in mS

uint16_t flashbytes;
char     pgwr_wait, bywr_wait, erase_wait;
char     bytebuff[256];

#define cmdanz   15
static const char PROGMEM cmds [cmdanz][9] =

//     1          2        3       4        5       6
  { "disable", "enable", "mosi", "miso", "reset", "clk",       \
//     7      8       9       10      11      12       13
    "rinv", "send", "read", "cmd3", "freq", "dump", "clkinv",  \
//    14      15
    "cmd4", "exit" };



char checkcmd(char *string)
{
  char  i,cp, ch;
  char  match;
  char  *s;

  i= 0;
  do
  {
    s= string;
    match= 1;
    cp= 0;
    do
    {
      ch= pgm_read_byte(&cmds[i][cp]);
      if (ch!= *s) match= 0;
//      printf("%d: %.2x = %c  %.2x = %c\n\r",cp,ch,ch,*s,*s);
      cp++; s++;
    }while(ch && match);
    if (match) return i+1;
    i++;
  } while (i < cmdanz);
  return 0;
}


void showactrstmode(void)
{
  crlf();
  uart_prints("actual mode is: ");
  if (rinv) uart_prints("invert "); else uart_prints("not invert ");
}

void showactclkmode(void)
{
  crlf();
  uart_prints("actual mode is: ");
  if (clkinv) uart_prints("invert "); else uart_prints("not invert ");
}

void get3cmdbytes(uint8_t *v1, uint8_t *v2, uint8_t *v3)
{
  crlf();
  uart_prints("Enter 3 instruction bytes: ");
  *v1 = gethex(1); uart_putchar(' ');
  *v2 = gethex(1); uart_putchar(' ');
  *v3 = gethex(1); uart_putchar(' ');
  crlf();
}

void get4cmdbytes(uint8_t *v1, uint8_t *v2, uint8_t *v3, uint8_t *v4)
{
  crlf();
  uart_prints("Enter 4 instruction bytes: ");
  *v1 = gethex(1); uart_putchar(' ');
  *v2 = gethex(1); uart_putchar(' ');
  *v3 = gethex(1); uart_putchar(' ');
  *v4 = gethex(1); uart_putchar(' ');
  crlf();
}

void send3cmdbytes(uint8_t v1, uint8_t v2, uint8_t v3)
{
  spi_out(v1);
  spi_out(v2);
  spi_out(v3);
}

uint8_t send4cmdbytes(uint8_t v1, uint8_t v2, uint8_t v3, uint8_t v4)
{
  uint8_t value;

  spi_out(v1);
  spi_out(v2);
  spi_out(v3);
  value= spi_out(v4);
  return value;
}

void spi_setfreq(uint8_t value)
{
  spi_disable();
  switch (value)
  {
    case 1 :   spi_init(3,0); break;          //    F_CPU / 128
    case 2 :   spi_init(2,0); break;          //    F_CPU / 64
    case 3 :   spi_init(2,1); break;          //    F_CPU / 32
    case 4 :   spi_init(1,0); break;          //    F_CPU / 16
    case 5 :   spi_init(1,1); break;          //    F_CPU / 8
    case 6 :   spi_init(0,0); break;          //    F_CPU / 4
    case 7 :   spi_init(0,1); break;          //    F_CPU / 2
    default : break;
  }
  aktfreq= value;
}

void dump(void)
{
  uint16_t startadr, stopadr;
  uint16_t cnt, pcnt;
  uint8_t buffer[8];
  uint8_t bcx, ch, i;

  uart_prints("\n\rStartadress: ");
  ch= gethex(1);
  startadr= (ch << 8);
  ch= gethex(1);
  startadr |= ch;

  uart_prints("\n\rStopadress : ");
  ch= gethex(1);
  stopadr= (ch << 8);
  ch= gethex(1);
  stopadr |= ch;
  crlf();
  crlf();
  cnt= startadr;
  pcnt= cnt;
  bcx= 0;
  for (i= 0; i<8; i++) buffer[i]= 0;
  do
  {
    send3cmdbytes(0x20, (cnt >> 8), (cnt & 0x00ff));

    buffer[bcx]= spi_read();
    cnt++; bcx++;
    if (bcx> 7)
    {
      printhex(pcnt >> 8); printhex(pcnt & 0x00ff);
      pcnt= cnt;
      uart_prints(": ");
      for (i= 0; i< 8; i++)
      {
        printhex(buffer[i]);
        uart_putchar(' ');
      }
      uart_prints("   ");
      for (i= 0; i< 8; i++)
      {
        if ((buffer[i]> 31) && (buffer[i]< 128))
          uart_putchar(buffer[i]);
        else
          uart_putchar('.');
      }
      crlf();
      bcx= 0;
    }
  } while(cnt < stopadr+1);
  if (bcx)
  {
    printhex(pcnt >> 8); printhex(pcnt & 0x00ff);
    uart_prints(": ");
    for (i= 0; i<8; i++)
    {
      if (i< bcx)
      {
        printhex(buffer[i]);
        uart_putchar(' ');
      }
      else
      {
        uart_prints("   ");
      }
    }
    uart_prints("   ");
    for (i= 0; i<8; i++)
    {
      if (i< bcx)
      {
        if ((buffer[i]> 31) && (buffer[i]< 128))
          uart_putchar(buffer[i]);
      }
    }
  }
  crlf();
}

void cmdinterpret(void)
{
  char ch;
  uint8_t b;
  uint8_t v1, v2, v3, v4;
  char inbuffer[13];

  uart_prints("\n\rSPI-evaluation\n\rexplore a SPI-Device\n\n\r");

  crlf();

  do
  {
    uart_prints("SPI:> ");
    readstr(&inbuffer[0],8);
    ch= checkcmd(&inbuffer[0]);
    switch(ch)
    {
      case 1 :                                // Hardware SPI disable
      {
        spi_disable();
        uart_prints("\n\rSPI disabled\n\r");
        break;
      }
      case 2 :                                // Hardware SPI enable
      {
        uart_prints("\n\rSPI enabled\n\r");
        spi_setfreq(aktfreq);
        break;
      }
      case 3 :                                // Mosi - Leitung
      {
        crlf();
        uart_prints("[0/1]: ");
        ch= uart_getchar();

        uart_putchar(ch);
        if (ch== '1') { spi_setmosi(); } else { spi_clrmosi(); }
        crlf();
        break;
      }
      case 4 :                                // Miso - Leitung lesen
      {
        crlf();
        if (spi_rdmiso()) {uart_prints("Miso= 1\n\r"); } else {uart_prints("Miso= 0\n\r");}
        break;
      }
      case 5 :                                // Reset - Leitung
      {
        crlf();
        uart_prints("[0/1]: ");
        ch= uart_getchar();
        uart_putchar(ch);
        if (ch== '1') { spi_setrst(); } else { spi_clrrst(); }
        crlf();
        break;
      }
      case 6 :                                // SCK (Clock) - Leitung
      {
        crlf();
        uart_prints("[0/1]: ");
        ch= uart_getchar();
        uart_putchar(ch);
        if (ch== '1') { spi_setclk(); } else { spi_clrclk(); }
        crlf();
        break;
      }
      case 7 :                               // Reset-Modus (invertiert oder nicht invertiert)
      {
        showactrstmode();
        uart_prints("[0/1]: ");
        ch= uart_getchar();
        uart_putchar(ch);
        if (ch== '1') {rinv= 1;} else {rinv= 0;}
        showactrstmode();
        crlf();
        break;
      }
      case 8 :                               // send Data-Byte
      {
        crlf();
        uart_prints("byte to send: ");
        b= gethex(1);
        spi_out(b);
        crlf();
        break;
      }
      case 9 :                               // read Data-Byte
      {
        crlf();
        uart_prints("readed byte: ");
        b= spi_read();
        printhex(b);
        crlf();
        break;
      }
      case 10 :                             // send 3-Byte instruction
      {
        get3cmdbytes(&v1, &v2, &v3);
        send3cmdbytes(v1, v2, v3);
        break;
      }
      case 11:                              // SPI clockfrequency
      {
        crlf();
        uart_prints("[1 .. 7]: ");
        ch= uart_getchar();
        uart_putchar(ch);
        spi_setfreq(ch-'0');
        crlf();
        break;
      }
      case 12:
      {
        dump();
        break;
      }
      case 13 :                               // Clock-Modus (invertiert oder nicht invertiert)
      {
        showactclkmode();
        uart_prints("[0/1]: ");
        ch= uart_getchar();
        uart_putchar(ch);
        if (ch== '1') {clkinv= 1;} else {clkinv= 0;}
        showactclkmode();
        spi_setfreq(aktfreq);
        crlf();
        break;
      }
      case 14 :                             // send 4-Byte instruction
      {
        get4cmdbytes(&v1, &v2, &v3, &v4);
        send4cmdbytes(v1, v2, v3, v4);
        break;
      }
      case 15 : uart_prints("\n\r"); break; // exit cmd-mode


/*      case 'x' :
        {
        } */
      default :
        {
          uart_putchar(ch);
          uart_prints("\n\runkown command\n\r");
          break;
        }
    }
  }while(ch != 15);
}

/*
   Reset Target und initialisiert program enable
   gibt angeschlossenen Controller zurueck
   ID-Nr:   73 = AT89S8253
            43 = AT89S4051
            23 = AT89S2051
            52 = AT89S52
            51 = AT89S51
            69 = ID defekt, MCU ist entweder 51 oder 52
*/

uint8_t progmode_init(void)
{
  uint8_t id;

  id= 0xff;
  // Chip reset
  clkinv= 0;
  spi_setfreq(aktfreq);
  spi_setrst();
  _delay_ms(15);
  spi_clrrst();
  _delay_ms(15);
  spi_setrst();
  _delay_ms(15);

  id= send4cmdbytes(0xac, 0x53, 0x00, 0x00);       // program enable command
  if (id== 0x69)
  {                                                // es ist ein S51 oder S52
    id= send4cmdbytes(0x28, 0x01, 0x00, 0x00);     // read ID command
    if (!(id == 0x51) && !(id |= 0x52))
    {
      id= 0x69;
      return id;
    }
    return id;
    uart_putchar('p');                             // sollte nie gesendet werden ...
  }
  id= 0xff;
  clkinv= 1;                                       // AT89S8253 und AT89Sx051 haben
  spi_setfreq(aktfreq);                            // invertierte Clockuebernahme
  spi_setrst();
  _delay_ms(15);
  spi_clrrst();
  _delay_ms(15);
  spi_setrst();
  _delay_ms(15);

  id= send4cmdbytes(0xac, 0x53, 0x00, 0x00);       // program enable command
  _delay_ms(4);
  id= send4cmdbytes(0x28, 0x00, 0x31, 0x00);       // read ID command S8253
  if (id== 0x73) return id;                        // es ist ein S8253
  id= send4cmdbytes(0x28, 0x00, 0x01, 0x00);       // read ID command Sx051
  if ((id== 0x43) || (id== 0x23)) return id;         // es ist ein Sx051

  return 0xff;
}

void mcu_readflash(void)
{
  uint8_t b, id;
  uint16_t cnt, proglen;

  id= progmode_init();

  switch (id)
  {
    case 0x69  : { proglen= 0x2000; break; }       // 89S51 oder S52, vorsichtshalber 8 kByte Flash
    case 0x51  : { proglen= 0x1000; break; }       // 89S51   =>  4 kByte Flash
    case 0x52  : { proglen= 0x2000; break; }       // 89S51   =>  8 kByte Flash
    case 0x73  : { proglen= 0x3000; break; }       // 89S8253 => 12 KByte Flash
    case 0x23  : { proglen= 0x0800; break; }       // 89S2051 =>  2 KByte Flash
    case 0x43  : { proglen= 0x1000; break; }       // 89S8253 =>  4 KByte Flash
    default: proglen= 0x0000; break;
  }

  printhex(proglen>> 8);
  printhex(proglen & 0x00ff);

  if (!proglen) return;

  for (cnt= 0; cnt< proglen; cnt++)
  {
    send3cmdbytes(0x20, (cnt >> 8), (cnt & 0x00ff));
    b= spi_read();
    printhex(b);
//    _delay_us(10);
  }

}


void mcu_reset(void)
{
  spi_clrrst();
  _delay_ms(15);
  spi_setrst();
  _delay_ms(15);
  spi_clrrst();
  _delay_ms(15);
  uart_putchar('r');
}

void mcu_erase(void)
{
  uint8_t b;

  uart_putchar('e');
  send4cmdbytes(0xac, 0x80, 0x00, 0x00);           // erase chip
  _delay_ms(e_delay);
}

void mcu_flash(void)
{
  uint16_t cx;
  uint8_t b, id;

  id= progmode_init();

  if ((id== 0x51) || (id== 0x52) || (id== 0x69))   // AT89S51 / S52
  {
    for (cx= 0; cx< flashbytes+1; cx++)
    {
      b= gethex(0);
      send4cmdbytes(0x40,cx >> 8, cx & 0x00ff,b);  // Bytewritemodus
      _delay_us(b_delay);
      printhex(b);                                 // Echo Hexzahl (Ascii)
    }
  }

  if (id== 0x73)                                   // AT89S8253
  {
    for (cx= 0; cx< flashbytes+1; cx++)
    {
      b= gethex(0);
      if ((cx % page8253) == 0)                       // Speichergroesse einer Page AT89S8253
      {
        _delay_us(p_delay);                           // Schreibdauer einer Page
        send4cmdbytes(0x50, cx>> 8, cx & 0x00ff,b);   // Pageadresse und Byte 0 der Page
      }
      else
      {
        spi_out(b);
      }
      printhex(b);
    }
    if (cx % page8253)                                // wenn eine Page nicht komplett ist
    {
      while(cx % page8253)                            // dann Speicherzellen mit 0xFF auffuellen
      {
        spi_out(0xff);
        cx++;
      }
    }
  }

  if ((id== 0x23) || (id== 0x43))                     // AT89S2051 / S4051
  {


//  nachfolgender Code funktioniert warum auch immer nicht: fix it !
/*
    for (cx= 0; cx< flashbytes+1; cx++)
    {
      b= gethex(0);
      if ((cx % page2051) == 0)                       // Speichergroesse einer Page AT89S2051
      {
        _delay_us(p_delay2);                          // Schreibdauer einer Page
        send4cmdbytes(0x50, cx>> 8, cx & 0x00ff,b);   // Pageadresse und Byte 0 der Page
      }
      else
      {
        spi_out(b);
      }
      printhex(b);
    }
    if (cx % page2051)                                // wenn eine Page nicht komplett ist
    {
      while(cx % page2051)                            // dann Speicherzellen mit 0xFF auffuellen
      {
        spi_out(0xff);
        cx++;
      }
    }
*/

    for (cx= 0; cx< flashbytes+1; cx++)
    {
      b= gethex(0);
      send4cmdbytes(0x40,cx >> 8, cx & 0x00ff,b);  // Bytewritemodus
      _delay_us(b_delay2);
      printhex(b);                                 // Echo Hexzahl (Ascii)
    }


  }
}


/* --------------------------------------------------

                       M A I N

   -------------------------------------------------- */

int main(void)
{
  char ch;
  uint8_t b;
  uint8_t v1, v2, v3, v4;
  char inbuffer[13];
  uint16_t i;

  uart_init(115200);

//  spi_init(2,1);
//  spi_disable();
  aktfreq= 2; clkinv= 0;
  spi_setfreq(aktfreq);

  while(1)
  {
    ch= uart_getchar();
    switch (ch)
    {
      case '1' :  mcutyp= progmode_init(); mcu_erase(); break;
      case '3' :  flashbytes= gethex16(1); break;               // Anzahl zu flashender Bytes
      case '4' :  mcu_reset(); break;
      case '5' :  cmdinterpret(); break;
      case '6' :  mcutyp= progmode_init(); mcu_flash(); break;
      case '7' :  mcu_readflash(); break;

      case '8' :
      {
        mcutyp= progmode_init();
        printhex(mcutyp);
        break;
      }

      case '9' :
      {
        printhex(0x18);
        break;
      }

      default : break;
    }
  }

  cmdinterpret();
  uart_prints("\nLeaving command mode..\n\r");


}

